﻿using System;
using System.Collections.Generic;
using System.Timers;
using System.Diagnostics;
using System.Threading;
using System.IO.Ports;

namespace Perfusjonssytem
{
    public enum ArduinoResult { Valid, Invalid, Disconnected, Timeout }

    public delegate void ConnectionDelegate();
    static class COM
    {
        static public event ConnectionDelegate ConnectedDel;
        static public event ConnectionDelegate DisconnectedDel;

        static SerialPort arduinoPort;
        const int BAUD_RATE = 115200;
        public const int FAILS_THRESHOLD = 3;
        const int TIMEOUT = 50;
        static int failsCount = 0;
        static bool connected;

        private static PerfusionForm perfusionFormCOM;

        static System.Timers.Timer connectionCheckTimer = new System.Timers.Timer(1000) { AutoReset = true, Enabled = false };

        public static SerialPort ArduinoPort { get => arduinoPort; set => arduinoPort = value; }
        public static int FailsCount { get => failsCount; set => failsCount = value; }
        public static bool Connected { get => connected; set => connected = value; }
        public static PerfusionForm PerfusionFormCOM { get => perfusionFormCOM; set => perfusionFormCOM = value; }

        static COM()
        {
            connectionCheckTimer.Elapsed += OnConnectionCheck;
        }

        static void OnConnectionCheck(object source, ElapsedEventArgs e)
        {
            if (ArduinoPort != null && !ArduinoPort.IsOpen) OnDisconnected();
        }

        static public string Encode(int channel, bool Open)
        {
            return string.Format("${0}/{1}#%", channel, Convert.ToInt32(Open));
        }

        static public ArduinoResult TrySending(string command)
        {
            ArduinoResult arduinoResult;
            while (FailsCount < FAILS_THRESHOLD)
            {
                arduinoResult = Send(command);

                if (arduinoResult != ArduinoResult.Invalid) // if send doesnt fail return the result, else try again up to 3 times
                {
                    return arduinoResult;
                }
            }
            FailsCount = 0;
            return ArduinoResult.Invalid;
        }

        //Send to serial port
        static public ArduinoResult Send(string command)
        {
            if (arduinoPort == null) return ArduinoResult.Disconnected;
            string response;
            try
            {
                arduinoPort.Write(command);
                response = arduinoPort.ReadLine();
                if (response.Contains("Valid"))
                {
                    return ArduinoResult.Valid;
                }
                else
                {
                    FailsCount++;
                    return ArduinoResult.Invalid;
                }
            }
            catch (TimeoutException)
            {
                FailsCount++;
                return ArduinoResult.Timeout;
            }
            catch (Exception)
            {
                return ArduinoResult.Disconnected;
            }
        }

        //Find port
        static private SerialPort FindAndConnect(int baudRate)
        {
            string[] ports = SerialPort.GetPortNames();
            SerialPort arduinoP = new SerialPort();
            arduinoP.ReadTimeout = TIMEOUT;
            arduinoP.BaudRate = baudRate;
            arduinoP.DtrEnable = true;
            arduinoP.RtsEnable = true;
            string pingResponse;

            foreach (string portName in ports)
            {
                arduinoP.PortName = portName;

                if (!arduinoP.IsOpen)
                {
                    try
                    {
                        arduinoP.Open();
                        arduinoP.Write("Connect%");
                        pingResponse = arduinoP.ReadLine();
                        if (pingResponse.Contains("Valid"))
                        {
                            // Return the serial port that responded "Valid"
                            OnConnected();
                            return arduinoP;
                        }
                    }
                    catch (Exception) { }

                    arduinoP.Close();
                }
            }

            // If no arduino was found
            OnDisconnected();
            return null;
        }

        //Connect to Arduino automatically
        static public void ConnectToArduino()
        {
            arduinoPort = FindAndConnect(BAUD_RATE);
        }

        static void OnConnected()
        {
            Connected = true;
            connectionCheckTimer.Enabled = true;

            if (ConnectedDel != null)
            {
                ConnectedDel();
            }
        }
        static void OnDisconnected()
        {
            Connected = false;
            connectionCheckTimer.Enabled = false;

            if (DisconnectedDel != null)
            {
                perfusionFormCOM.Invoke(DisconnectedDel);
            }
        }

    } //End of COM class
}