﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Timers;

namespace Perfusjonssytem
{
    public delegate void SequenceDelegate(int index);
    public delegate void UnlockDelegate();
    public delegate void ResultDelegate(ArduinoResult arduinoResult);
    class Sequence
    {
        public SequenceDelegate actionStartedDelegate;
        public SequenceDelegate actionInvalidDelegate;
        public UnlockDelegate unlockDelegate;
        public ResultDelegate resultDelegate;

        object stopSync;
        public object runningSync;

        PerfusionForm perfusionForm;
        bool stopCommand;
        bool wasStoped;
        private bool running;
        List<PAction> toDoActionList;
        List<PAction> completedActionList;
        DateTime timeStarted;
        Stopwatch stopwatch;
        System.Timers.Timer sequenceTickTimer;
        const int TICK_TIMER_INTERVAL = 25;
        EventWaitHandle waitForActionEwh;
        PAction currentAction;

        ArduinoResult arduinoResult;

        LogData logData;

        Thread sequenceThread;

        public Sequence(List<PAction> actionList, PerfusionForm perfustionFrom_)
        {
            ToDoActionList = CopyList(actionList);
            perfusionForm = perfustionFrom_;

            completedActionList = new List<PAction>();

            stopSync = new object();
            runningSync = new object();

            stopwatch = new Stopwatch();
            sequenceTickTimer = new System.Timers.Timer() { Interval = TICK_TIMER_INTERVAL };
            sequenceTickTimer.Elapsed += SequenceTickTimer_Tick;
            waitForActionEwh = new EventWaitHandle(false, EventResetMode.AutoReset);
        }

        private List<PAction> CopyList(List<PAction> actionList)
        {
            List<PAction> actionBuffer = new List<PAction>();
            for (int i = 0; i < actionList.Count; i++)
            {
                actionBuffer.Add(actionList[i]);
            }
            return actionBuffer;
        }

        public bool StopCommand { get => stopCommand; set => stopCommand = value; }
        public bool WasStoped { get => wasStoped; set => wasStoped = value; }
        public DateTime TimeStarted { get => timeStarted; set => timeStarted = value; }
        internal List<PAction> ToDoActionList { get => toDoActionList; set => toDoActionList = value; }
        internal List<PAction> CompletedActionList { get => completedActionList; set => completedActionList = value; }
        public bool Running { get => running; set => running = value; }
        internal LogData LogData { get => logData; set => logData = value; }

        bool bufferStop;
        void SequenceTickTimer_Tick(object sender, EventArgs e)
        {
            lock (stopSync) bufferStop = stopCommand;
            
            if (stopwatch.ElapsedMilliseconds >= currentAction.Duration - (TICK_TIMER_INTERVAL / 2) || bufferStop)
            {
                waitForActionEwh.Set();
            }
        }

        public void StartSequenceThread()
        {
            sequenceThread = new Thread(new ThreadStart(StartSequence));
            sequenceThread.Start();
        }

        bool formExit;
        void StartSequence() // send the whole sequence one by one to the arduino via the send method in the COM class
        {
            lock (runningSync)
            {
                Running = true;
            }
            TimeStarted = DateTime.Now;
            for (int i = 0; i < ToDoActionList.Count; i++) // loops through all actions
            {
                currentAction = ToDoActionList[i];

                stopwatch.Reset();
                arduinoResult = COM.TrySending(COM.Encode(ToDoActionList[i].ChannelNum,true));

                lock (perfusionForm.ExitingSync) formExit = perfusionForm.Exiting;

                if (arduinoResult != ArduinoResult.Valid) // unsuccessful after 3  attempts
                {
                    stopwatch.Reset();
                    if (!formExit) perfusionForm.Invoke(resultDelegate, arduinoResult);
                    break; // ends the sequence
                }

                if (!formExit) perfusionForm.Invoke(actionStartedDelegate, i);
                stopwatch.Start();
                sequenceTickTimer.Start();

                waitForActionEwh.WaitOne(); // Waiting for the current action to be completed, or the sequence to be stoped
                stopwatch.Stop();
                sequenceTickTimer.Stop();

                // Add the stopped action
                if (stopCommand)
                {
                    PAction stoppedAction = new PAction(ToDoActionList[i].Substance, ToDoActionList[i].ChannelNum, (int)stopwatch.ElapsedMilliseconds);
                    CompletedActionList.Add(stoppedAction);
                }
                // Add the completed action
                else
                {
                    CompletedActionList.Add(ToDoActionList[i]);
                }

                arduinoResult = COM.TrySending(COM.Encode(ToDoActionList[i].ChannelNum, false));

                lock (perfusionForm.ExitingSync) formExit = perfusionForm.Exiting;

                if (arduinoResult != ArduinoResult.Valid) // unsuccessful after 3 atempts
                {
                    if (!formExit)
                    {
                        perfusionForm.Invoke(actionInvalidDelegate, i);
                        perfusionForm.Invoke(resultDelegate, arduinoResult);
                    }
                    break; // ends the sequence
                }

                if (stopCommand) break;
                
            } // loops through actions

            //colors all actions green if the sequence finished succesfully
            if (!stopCommand && arduinoResult == ArduinoResult.Valid) perfusionForm.Invoke(actionStartedDelegate, perfusionForm.SequenceForm.ActionList.Count);


            CreateLog();

            lock (runningSync) Running = false;

            // Dont modify GUI if the form is closing to avoid crash
            
            lock (perfusionForm.ExitingSync) formExit = perfusionForm.Exiting;

            if (!formExit)
            {
                perfusionForm.Invoke(unlockDelegate);
            }
        }

        

        public void WaitForSequenceThread()
        {
            sequenceThread.Join();
        }
        public void StopSequence()
        {
            lock (stopSync)
            {
                StopCommand = true;
            }
        }

        void CreateLog()
        {
            Result resultToLog;

            if (arduinoResult != ArduinoResult.Valid)
            {
                resultToLog = Result.Disconnected;
            }
            else if (stopCommand)
            {
                resultToLog = Result.Stoped;
            }
            else
            {
                resultToLog = Result.Completed;
            }

            logData = new LogData(TimeStarted, CompletedActionList, resultToLog, "");
        }
    }

    struct PAction
    {
        string substance;
        int channelNum;
        int duration;

        public string Substance { get => substance; set => substance = value; }
        public int ChannelNum { get => channelNum; set => channelNum = value; }
        public int Duration { get => duration; set => duration = value; }

        public PAction(string substance_, int channelNum_, int duration_)
        {
            substance = substance_;
            channelNum = channelNum_;
            duration = duration_;
        }

    }
}
