﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Perfusjonssytem
{
    public partial class SequenceForm : Form
    {
        List<PAction> actionList;
        BindingList<Channel> seqChannels;
        Sequence currentSequence;
        PerfusionForm perfusionForm;

        Color actionHighlightColor;
        Color actionCompletedColor;
        Color actionStopedColor;

        internal List<PAction> ActionList { get => actionList; }
        internal Sequence CurrentSequence { get => currentSequence; set => currentSequence = value; }
        internal PictureBox PbArduinoStatusSeq { get => pbArduinoStatusSeq; set => pbArduinoStatusSeq = value; }
        internal BindingList<Channel> SeqChannels { get => seqChannels; set => seqChannels = value; }

        public SequenceForm(PerfusionForm pf)
        {
            InitializeComponent();
            this.TopLevel = false;

            perfusionForm = pf;

            actionList = new List<PAction>();
            seqChannels = new BindingList<Channel>();

            actionHighlightColor = Color.FromArgb(0, 113, 175);
            actionCompletedColor = Color.FromArgb(24, 89, 39);
            actionStopedColor = Color.FromArgb(196, 8, 8);

            cbChannelSeq.DataSource = seqChannels;
            COM.ConnectedDel += UpdatebtnStartSequence;
            COM.DisconnectedDel += UpdatebtnStartSequence;
            COM.DisconnectedDel += StopSequenceOnDisconnect;

            Settings.ReadPreviousSequence(actionList);
            UpdateGUI();
        }

        private void btnAdd_Click(object sender, EventArgs e)
        {
            Channel ch = (Channel)cbChannelSeq.SelectedItem;
            ActionList.Add(new PAction(ch.Substance, ch.ChannelNum, ((int)Math.Round(((nudTimeSec.Value * 1000 + nudTimeMin.Value * 60000) / 10))) * 10));

            btnSaveLog.Enabled = false;
            UpdateGUI();
        }
        private void btnStartSequence_Click(object sender, EventArgs e)
        {
            ResetActionLVHighlight();
            currentSequence = new Sequence(ActionList, perfusionForm);
            currentSequence.actionStartedDelegate += IncrementActionLVHighlight;
            currentSequence.unlockDelegate += UnlockGUISeq;
            currentSequence.actionInvalidDelegate += SequenceError;
            currentSequence.resultDelegate += DisplaySequenceError;
            LockGUISeq();
            perfusionForm.SeqIndicator.Start();
            perfusionForm.seqDurationTimer.Start();
            perfusionForm.SeqDurationSW.Restart();
            btnStopSequence.Enabled = true;
            lvActions.SelectedIndices.Clear();

            currentSequence.StartSequenceThread();
        }
        private void btnRemove_Click(object sender, EventArgs e)
        {
            if (lvActions.SelectedIndices.Count != 0)
            {
                ActionList.RemoveAt(lvActions.SelectedIndices[0]);

                btnSaveLog.Enabled = false;
                UpdateGUI();
            }
        }
        private void btnStopSequence_Click(object sender, EventArgs e)
        {
            if (currentSequence != null) currentSequence.StopSequence();
            perfusionForm.SeqIndicator.Stop();
            perfusionForm.StopSequenceDuration();

            btnStopSequence.Enabled = false;
        }

        private void StopActionLVHighlight(object sender, EventArgs e)
        {
            int currentIndex = 0;
            lvActions.Items[currentIndex].BackColor = actionStopedColor;
        }
        private void SequenceError(int failedIndex)
        {
            lvActions.Items[failedIndex].BackColor = Color.Red;
            perfusionForm.SeqIndicator.Stop();
            perfusionForm.StopSequenceDuration();
        }
        private void DisplaySequenceError(ArduinoResult result)
        {
            string errorMessage = "The sequence ran into an error.\n";

            switch (result)
            {
                case ArduinoResult.Invalid:
                    errorMessage += "The controller was unable to execute the given command, please restart it.";
                    break;
                case ArduinoResult.Disconnected:
                    errorMessage += "The connection to the controller was lost, please reconnect.";
                    break;
                case ArduinoResult.Timeout:
                    errorMessage += "The controller is connected but unresponsive, please check that the control mode switch is in the correct position. If so, try restarting the controller.";
                    break;
                default:
                    break;
            }
            MessageBox.Show(errorMessage, "Controller Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
        private void nudTime_ValueChanged(object sender, EventArgs e)
        {
            UpdatebtnAdd();
        }
        private void lvActions_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdatebtnRemove();
        }
        private void StopSequenceOnDisconnect()
        {
            if (currentSequence != null) currentSequence.StopSequence();
        }
        private void UpdatelvActions()
        {
            lvActions.Items.Clear();

            perfusionForm.SeqIndicator.ClearSegments();

            for (int i = 0; i < ActionList.Count; i++)
            {
                string duration;
                if (ActionList[i].Duration < 60000) duration = (ActionList[i].Duration / 1000.0).ToString() + "s";
                else duration = (ActionList[i].Duration / 60000).ToString() + "m " + ((ActionList[i].Duration % 60000) / 1000.0).ToString() + "s";
                string[] aStr =
                {
                    ActionList[i].ChannelNum.ToString(),
                    ActionList[i].Substance,
                    duration
                };
                ListViewItem item = new ListViewItem(aStr);
                lvActions.Items.Add(item);
                perfusionForm.SeqIndicator.AddSegment(ActionList[i].Substance, ActionList[i].Duration);
            }

        }
        private void UpdateActionList()
        {
            ActionList.RemoveAll(FindNoneActions);
        }
        private bool FindNoneActions(PAction p)
        {
            if (p.Substance == "None")
            {
                return true;
            }
            return false;
        }
        public void UpdateGUI()
        {
            UpdateActionListNames();
            UpdateActionList();
            UpdatelvActions();
            UpdatebtnRemove();
            UpdatebtnStartSequence();
            UpdateChannelLists();
            UpdatebtnClear();
        }
        private void UpdatebtnStartSequence()
        {
            if (actionList.Count != 0 && COM.Connected)
            {
                btnStartSequence.Enabled = true;
            }
            else
            {
                btnStartSequence.Enabled = false;
            }
        }
        private void UpdateChannelLists()
        {
            Channel selectedChannel = new Channel(0, "Wait");

            if (cbChannelSeq.SelectedItem != null)
            {
                // remember the channel that was selected
                selectedChannel = (Channel)cbChannelSeq.SelectedItem;
            }

            // clear and add new channels from cbSubstanceList
            seqChannels.Clear();
            seqChannels.Add(new Channel(0, "Wait"));
            for (int i = 0; i < perfusionForm.SettingsForm.CbSubstanceList.Count; i++)
            {
                if (perfusionForm.SettingsForm.CbSubstanceList[i].SelectedItem != null && perfusionForm.SettingsForm.CbSubstanceList[i].SelectedItem.ToString() != "None")
                {
                    seqChannels.Add(new Channel(i + 1, perfusionForm.SettingsForm.CbSubstanceList[i].SelectedItem.ToString()));
                }
            }

            // reselect the previously selected channel if it still exists
            for (int i = 0; i < cbChannelSeq.Items.Count; i++)
            {
                if (((Channel)cbChannelSeq.Items[i]) == selectedChannel)
                {
                    cbChannelSeq.SelectedIndex = i;
                    break;
                }
            }
        }
        private void UpdateActionListNames()
        {
            string substanceInCB;
            for (int i = 0; i < ActionList.Count; i++)
            {
                if (ActionList[i].ChannelNum != 0)
                {
                    substanceInCB = perfusionForm.SettingsForm.CbSubstanceList[ActionList[i].ChannelNum - 1].SelectedItem.ToString();
                }
                else
                {
                    substanceInCB = "Wait";
                }

                if (ActionList[i].Substance != substanceInCB) // checks if the name of the substance has been changed, and if so change it in the actionlist accordingly
                {
                    ActionList[i] = new PAction(substanceInCB, ActionList[i].ChannelNum, ActionList[i].Duration);
                }
            }
        }
        public void ResetActionLVHighlight()
        {
            for (int i = 0; i < lvActions.Items.Count; i++)
            {
                lvActions.Items[i].BackColor = ListView.DefaultBackColor;
            }
        }
        private void IncrementActionLVHighlight(int currentIndex)
        {
            for (int i = 0; i < lvActions.Items.Count; i++)
            {
                if (currentIndex > i) lvActions.Items[i].BackColor = actionCompletedColor;
                else if (currentIndex == i) lvActions.Items[i].BackColor = actionHighlightColor;
            }
        }
        private void LockGUISeq()
        {
            btnAdd.Enabled = false;
            btnRemove.Enabled = false;
            txtComment.Enabled = false;
            nudTimeMin.Enabled = false;
            nudTimeSec.Enabled = false;
            btnStartSequence.Enabled = false;
            cbChannelSeq.Enabled = false;
            btnClear.Enabled = false;
            btnSaveLog.Enabled = false;
            btnMoveActionUp.Enabled = false;
            btnMoveActionDown.Enabled = false;

            perfusionForm.BtnManual.Enabled = false;
            perfusionForm.BtnSettings.Enabled = false;
        }
        public void UnlockGUISeq()
        {
            lvActions.Enabled = true;
            btnAdd.Enabled = true;
            txtComment.Enabled = true;
            nudTimeMin.Enabled = true;
            nudTimeSec.Enabled = true;
            cbChannelSeq.Enabled = true;
            if (COM.Connected) btnStartSequence.Enabled = true;
            btnClear.Enabled = true;
            btnStopSequence.Enabled = false;
            btnSaveLog.Enabled = true;
            btnMoveActionUp.Enabled = true;
            btnMoveActionDown.Enabled = true;

            perfusionForm.BtnManual.Enabled = true;
            perfusionForm.BtnSettings.Enabled = true;
            perfusionForm.StopSequenceDuration();
        }
        private void UpdatebtnRemove()
        {
            if (lvActions.SelectedIndices.Count != 0 && (currentSequence == null || !currentSequence.Running))
            {
                btnRemove.Enabled = true;
            }
            else
            {
                btnRemove.Enabled = false;
            }
        }
        private void UpdatebtnClear()
        {
            if (lvActions.Items.Count != 0)
            {
                btnClear.Enabled = true;
            }
            else
            {
                btnClear.Enabled = false;
            }
        }
        private void UpdatebtnAdd()
        {
            if (nudTimeSec.Value == 0) btnAdd.Enabled = false;
            else btnAdd.Enabled = true;
        }
        private void btnClear_Click(object sender, EventArgs e)
        {
            actionList.Clear();

            perfusionForm.SeqIndicator.ClearSegments();

            btnSaveLog.Enabled = false;
            UpdateGUI();
        }

        private void SequenceForm_Click(object sender, EventArgs e)
        {
            lvActions.SelectedIndices.Clear();
        }

        private void btnSaveLog_Click(object sender, EventArgs e)
        {
            if (Log.LogFilePathFolder != null && Log.LogFilePathFolder.Contains(AppDomain.CurrentDomain.BaseDirectory + "Logs"))
            {
                if (currentSequence != null)
                {
                    currentSequence.LogData.Comment = txtComment.Text;
                    Log.SaveToFile(currentSequence.LogData);
                    btnSaveLog.Enabled = false;
                }
            }
            else MessageBox.Show("Unable to save log due to an invalid folder location.", "Unable to save log", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void btnMoveActionUp_Click(object sender, EventArgs e)
        {
            if (lvActions.SelectedItems.Count > 0)
            {
                int index = lvActions.SelectedIndices[0];

                if (index >= 1)
                {
                    PAction p = ActionList[index];
                    ActionList.RemoveAt(index);
                    ActionList.Insert(index - 1, p);
                    UpdateGUI();
                    lvActions.Items[index - 1].Selected = true;
                }
            }
        }

        private void btnMoveActionDown_Click(object sender, EventArgs e)
        {
            if (lvActions.SelectedItems.Count > 0)
            {
                int index = lvActions.SelectedIndices[0];

                if (index <= lvActions.Items.Count - 2)
                {
                    PAction p = ActionList[index];
                    ActionList.RemoveAt(index);
                    ActionList.Insert(index + 1, p);
                    UpdateGUI();
                    lvActions.Items[index + 1].Selected = true;
                }
            }
        }
    }
}
